/*
 	Copyright (C) 2001, 2002 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <string.h>
#include <limits.h>
#include <modules.h>
#include <util.h>
#include <lists.h>
#include <dstring.h>
#include <stdlib.h>
#include "version.h"
//// \author[Artem V. Andreev]
//// \copyright[2001, 2002 Artem V. Andreev -- See \ref[sl(1)] for details]
//// \section[NAME]
//// \em[fpmath] -- a module for Tensile
//// \section[DESCRIPTION]
//// \em[fpmath] adds floating-point support to Tensile.

static module *this_module;
extern procstack *get_pstack(void);
extern variable *safe_get_var(interpreter *i, char *name, int writing);

#define INT_LENGTH 0

static double convert_to_d(char *str)
{
	char *check = "";
	double nv = strtod(str, &check);
	if(*check)
		nv = (double)convert_to_int(str);
	return nv;
}

void fpmath_LTX_init(module *cur_mod)
{
	this_module = cur_mod;
}
//// \section[FUNCTIONS]

//// isfloat(num)
/// \returns true if \var[num] is a legal floating-point number
static void bi_isfloat(chunk *current, interpreter *i, value *args)
{
	char *str = DS_BODY(args->val);
	char *check = "";
	strtod(str, &check);
	if(!*check)
		ds_appendstr(get_pstack()->state->buffer, "true");
}

/// fadd(num1 num2 ...)
/// \returns the sum of all its argument
static void bi_fadd(chunk *current, interpreter *i, value *args)
{
	double value = 0;

	for(; args; args = list_next(args))
		value += convert_to_d(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromdouble(value));
}

//// fpostincr(var val)
/// adds \var[val] to \var[var]
/// \returns old value of \var[var]
static void bi_fpostincr(chunk *current, interpreter *i, value *args)
{
	double val = 0;
   	double ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ds_append(get_pstack()->state->buffer, var->vals->val);
	val = convert_to_d(DS_BODY(next_value(args)->val));
	ovalue = convert_to_d(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromdouble(val + ovalue));
}

//// fincr(var val)
/// adds \var[val] to \var[var]
/// \returns new value of \var[var]
static void bi_fincr(chunk *current, interpreter *i, value *args)
{
	double val = 0; 
	double ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	val = convert_to_d(DS_BODY(next_value(args)->val));
	ovalue = convert_to_d(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromdouble(val + ovalue));
	ds_append(get_pstack()->state->buffer, var->vals->val);
}

//// fpostdecr(var val)
/// subtracts \var[val] from \var[var]
/// \returns old value of \var[var]
static void bi_fpostdecr(chunk *current, interpreter *i, value *args)
{
	double val = 0;
   	double ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ds_append(get_pstack()->state->buffer, var->vals->val);
	val = convert_to_d(DS_BODY(next_value(args)->val));
	ovalue = convert_to_d(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromdouble(ovalue - val));
}

//// fdecr(var val)
/// subtracts \var[val] from \var[var]
/// \returns old value of \var[var]
static void bi_fdecr(chunk *current, interpreter *i, value *args)
{
	double val = 0.0;
	double ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ovalue = convert_to_d(DS_BODY(var->vals->val));
	val = convert_to_d(DS_BODY(next_value(args)->val));
	ds_assign(var->vals->val, ds_fromdouble(ovalue - val));
	ds_append(get_pstack()->state->buffer, var->vals->val);
}

/// fmul(val1 val2 ...)
/// \returns a multiply of \var[val]s
static void bi_fmul(chunk *current, interpreter *i, value *args)
{
	double value = 1.0;

	for(; args; args = list_next(args))
		value *= convert_to_d(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromdouble(value));
}

/// fsub(val1 val2)
//// \returns difference between \em[val1] and \em[val2]
static void bi_fsub(chunk *current, interpreter *i, value *args)
{
	double v1 = convert_to_d(DS_BODY(args->val));
	double v2 = convert_to_d(DS_BODY(((value	*)list_next(args))->val));
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v1 - v2));
}

/// fdiv(val1 val2)
//// \returns quotient of \em[val1] and \em[val2]
static void bi_fdiv(chunk *current, interpreter *i, value *args)
{
	double v1 = convert_to_d(DS_BODY(args->val));
	double v2 = convert_to_d(DS_BODY(((value	*)list_next(args))->val));
	if(!v2)
		recover_error("zero division");
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v1 / v2));
}

/// fneg(val)
//// \returns a negated \var[val]
static void bi_fneg(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromdouble(-v));
}

/// fabs(val)
/// \returns the absolute value of \var[val]
static void bi_fabs(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromdouble(fabs(v)));
}

/// fint(val)
/// \returns integral part of \var[val]
static void bi_fint(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	modf(v, &v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

/// frac(val)
/// \returns fractional part of \var[val]
static void bi_frac(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val)), ip;
	v = modf(v, &ip);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

/// fsqrt(val)
//// \returns a square root of \var[val]
static void bi_fsqrt(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	v = sqrt(v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

/// fln(val)
//// \returns a natural logarithm of \var[val]
static void bi_fln(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	v = log(v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

//// fsin(val)
//// \returns a sine of \var[val] (in radians)
static void bi_fsin(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	v = sin(v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

/// fcos(val)
/// \returns a cosine of \var[val] (in radians)
static void bi_fcos(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	v = cos(v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

//// fexp(val)
/// \returns \em[e]^\var[val]
static void bi_fexp(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	v = exp(v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

//// fpower(val1 val2)
//// \returns \var[val1]^\var[val2]
static void bi_fpower(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val)), p = convert_to_d(DS_BODY(next_value(args)->val));
	v = pow(v, p);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

//// fprecision([prec])
/// if \var[prec] is present, sets the floating point precision to it
/// \returns the old value of precision
static void bi_fprecision(chunk *current, interpreter *i, value *args)
{
	int prec;
	if(args)
		prec = ds_dbl_set_precision(convert_to_int(DS_BODY(args->val)));
	else
		prec = ds_dbl_set_precision(-1);
	ds_append(get_pstack()->state->buffer, ds_fromint(prec, 10, 0));
}

//// Comparison functions
/// \em[eq], \em[neq], \em[less], \em[greater], \em[le], \em[ge]
/// accept \em[float] as their first argument, resulting in floating
//// point comparison
static void pass_thru(chunk *current, interpreter *i, value *args, int opcode)
{
	instruction_t *old_vector = get_ma_data(this_module, i);
	if(old_vector)
		old_vector[opcode](current, i, args);
}

enum opcodes_in_question { op_eq = 0, op_neq, op_less, op_greater, op_le, op_ge };
// (refer to the following instruction table)

static void bi_eq(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_eq);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val))  == convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_neq(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_neq);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val)) != convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_less(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_less);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val)) < convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_greater(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_greater);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val)) > convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_le(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_le);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val)) <= convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_ge(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_ge);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val)) >= convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static master_instruction_src fpmath_add_builtins[] = {
							{"isfloat", 1, bi_isfloat, 1, 1, 1, 1},
							{"fadd", 1, bi_fadd, 2, INT_MAX, 2, 1},
							{"fsub", 1, bi_fsub, 2, 2, 2, 1},
							{"fmul", 1, bi_fmul, 2, INT_MAX, 2, 1},
							{"fdiv", 1, bi_fdiv, 2, 2, 2, 1},
							{"fneg", 1, bi_fneg, 1, 1, 1, 1},
							{"fabs", 1, bi_fabs, 1, 1, 1, 1},
							{"fincr", 1, bi_fincr, 2, 2, 2},
							{"fdecr", 1, bi_fdecr, 2, 2, 2},
							{"fpostincr", 1, bi_fpostincr, 2, 2, 2},
							{"fpostdecr", 1, bi_fpostdecr, 2, 2, 2},
							{"fint", 1, bi_fint, 1, 1, 1, 1},
							{"frac", 1, bi_frac, 1, 1, 1, 1},
							{"fprecision", 1, bi_fprecision, 0, 1, 1},
							{"fln", 1, bi_fln, 1, 1, 1},
							{"fsqrt", 1, bi_fsqrt, 1, 1, 1},
							{"fsin", 1, bi_fsin, 1, 1, 1},
							{"fcos", 1, bi_fcos, 1, 1, 1},
							{"fexp", 1, bi_fexp, 1, 1, 1},
							{"fpower", 1, bi_fpower, 2, 2, 2},
							{NULL}
							};


static master_instruction_src fpmath_builtins[] = {
							{"eq", 1, bi_eq, 2, 3, 2, 1},
							{"neq", 1, bi_neq, 2, 3, 2, 1},
							{"less", 1, bi_less, 2, 3, 2, 1},
							{"greater", 1, bi_greater, 2, 3, 2, 1},
							{"le", 1, bi_le, 2, 3, 2, 1},
							{"ge", 1, bi_ge, 2, 3, 2, 1},
							{NULL}
};

typedef struct fpmath_data
{
	instruction_t old[NUMOFITEMS(fpmath_builtins)];
	int precision;
} fpmath_data;

void fpmath_LTX_attach(mod_attach *ma)
{
	 ma->data = xmalloc(sizeof(fpmath_data));
	 append_builtins(ma->i, fpmath_add_builtins);
	 replace_builtins(ma->i, fpmath_builtins, ((fpmath_data *)ma->data)->old);
}

const char fpmath_LTX_version[] = VERSION;
const char fpmath_LTX_date[] = __DATE__;
const char fpmath_LTX_srcdate[] = "$Date: 2002/12/09 10:41:21 $";

//// \trailer
